/*
 * Hanashizume no Matsuri for X
 * Copyright (c) 2000 TF <tf@denpa.org>
 * All rights reserverd.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgment:
 *      This product includes software developed by TF.
 * 4. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: hana.h,v 1.38 2000/11/27 11:26:52 tf Exp $
 */
#ifndef __INCLUDE_HANA_H
#define __INCLUDE_HANA_H

#include <stdio.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xlocale.h>
#include <X11/Shell.h>

#include "image.h"
#include "filebrowser.h"

#ifdef DEBUG
#define	DPRINTF(arg)	printf arg
#else
#define	DPRINTF(arg)
#endif

#define BUFLEN 512
#define PATHLEN 256
#define NAMELEN 16

#ifndef MAX_PATH
#define MAX_PATH 1024
#endif

/*
 *  command
 */
#define NONE        0x0
#define PRINT_TEXT  0x1
#define LINE_FEED   0x2
#define SET_CG      0x3
#define SET_CG_FADE 0x4
#define WAIT_CLICK  0x5
#define WAIT        0x6
#define NEW_PAGE    0x7
#define ALTERNATIVE 0x8
#define PLAY_MUSIC  0x9
#define FLASH       0xa
#define COLOR_MASK  0xb
#define HIGHLIGHT   0xc
#define FADE_OUT    0xd
#define SET_CG_OMAKE 0xe

/*
 * states of display engine
 */
#define STATE_PROC 0x0
#define STATE_WAIT 0x1
#define STATE_ALTN 0x2
#define STATE_CG   0x3

/*
 * states of scenario engine
 */
#define STAT_INIT 0x0
#define STAT_PROC 0x1
#define STAT_ALTN 0x2
#define STAT_WAIT 0x3
#define STAT_OPENING 0x4
#define STAT_OPENING_FAST 0x5
#define STAT_ENDING 0x6
#define STAT_ENDING_FAST 0x7
#define STAT_OMAKE 0x8

#define SPEED_SLOW   0x0
#define SPEED_NORMAL 0x1
#define SPEED_FAST   0x2
#define SPEED_VFAST  0x3

#define MUSIC_NONE 0x0
#define MUSIC_CDDA 0x1
#define MUSIC_MIDI 0x2
#define MUSIC_MP3  0x3
#define NMUSIC     0x4

/* constant for image processing */
#define BG_DENSITY 26 /* 32 * 0.8 = 20% dark */
#define MAX_DENSITY 32

/* max number of alternative */
#define MAX_ALT 4
#define MARGIN_X_OPENING 2
#define MARGIN_Y_OPENING 8

/* location of the title screen in the script */
#define TITLE_OFFSET 1600

#ifndef ORIGDIR
#define ORIGDIR "/please/set/origdir"
#endif

#ifndef CDDEVICE
#define CDDEVICE "/please/set/cddevice"
#endif

#ifndef MIDIPLAYER
#define MIDIPLAYER "/please/set/midiplayer"
#endif

#ifndef MP3PLAYER
#define MP3PLAYER "/please/set/mp3player"
#endif

#ifndef MP3DIR
#define MP3DIR "/please/set/mp3dir"
#endif

#ifndef MP3FILENAME
#define MP3FILENAME "please_set_mp3filename"
#endif

typedef struct {
  Bool enable;
  int r;
  int g;
  int b;
} GRAYSCALE;

typedef struct {
  GRAYSCALE grayscale;
  Bool fade;
  int fadetime;
} ImageAttribute;

typedef struct {
  int type;
  int ubuf[1024]; /* buffer for UCS-2 string */
  unsigned char buf[512];
  int arg;
  ImageAttribute attr;
} COMMAND;

typedef struct RESINFO {
  int type;
  char name[NAMELEN];
  int lang;
  long offset;
  struct RESINFO *next;
} RESINFO;

typedef struct {
  FILE *fd;
  RESINFO *rlist;
} ARCHIVE;

typedef struct {
  char orig[PATHLEN];
  char cddevice[PATHLEN];
  char midiwork[PATHLEN];
  char midiplayer[PATHLEN];
  char mp3player[PATHLEN];
  char mp3dir[PATHLEN];
  char mp3filename[PATHLEN];
  char savedir[PATHLEN];
  char imagedir[PATHLEN];
  char globalinfo[PATHLEN];
} PATH;

typedef struct {
  /* location in scenario */
  char script[NAMELEN];
  int block;
  int offset;
  char cgname[NAMELEN];

  /* music status */
  int track;

  /* location in scenario at last page end(used to save/load) */
  char pe_script[NAMELEN];
  int pe_block;
  int pe_offset;
  int pe_init_state;
  int pe_alternative;
  
  /* internal state */
  int state;
  int init_state; /* script $B$N=i4|>uBV$GA*Br;h$+K\J8$+$r;XDj(B */
  int clear;
  Bool newpage;

  /* flags */
  int flag_megumi;
  int flag_yue;
  int flag_ruria;
  int flag_sarasa;
  int flag_hinata;
  int flag_himiko;
  int flag_mao;
  int flag_kurumi;
  int flag_minako;

  /* for alternative */
  Bool alternative;
  int alt_num;
  int alt[MAX_ALT][BUFLEN];
  int alt_val;
  Bool alt_opening;

  Image *image;	/* XXX for save image */
} SCN_STAT;

typedef struct {
  Bool skip;
  int speed;
  Bool cg_fade;

  /* music */
  int music_mode;
  int music_fd;
  Bool music_init[NMUSIC];
} SYS_STAT;

/* main.c */
extern int open_archives(void);
extern int close_archives(void);
extern void quit_proc(Widget, XtPointer, XtPointer);

/* image.c */
extern void set_ximage(Image *, int, int, int, GRAYSCALE);
extern void mix_ximages(Image *, Image *,int width, int height, int, int, GRAYSCALE);
extern Image *ImageCreate(Display *, Window, int, int);
extern void ImagePut(Image *);
extern void ImageDestroy(Image *);
extern void ImageAttributeInit(ImageAttribute *);
extern void SetWhite(Image *, int, int);
extern void InitImageProcTable(void);

/* jpeg.c */
extern unsigned char *get_jpeg(FILE *, long, int, int);

/* resinfo.c */
extern int scan_dll(FILE *fp, RESINFO *);
extern void add_resinfo(RESINFO *);
extern RESINFO *create_rlist(void);
extern void free_rlist(RESINFO *);
extern void print_rlist(RESINFO *);
extern RESINFO *search_rlist(RESINFO *, const char *);

/* scenario.c */
extern void init_scenario(void);
extern void free_script(void);
extern void do_scenario(ARCHIVE, COMMAND *);
extern void parse_block(int, COMMAND *);
extern int proc_alternative(int);

/* unicode.c */
extern XChar2b u2xch(int);
extern int make_u2e_table(void);
extern int puteuc_string(const int *);

/* music.c */
extern void music_init(void);
extern void music_term(void);
extern void music_play(int, Bool);
extern void music_stop(void);

/* music_cd.c */
extern int music_cd_process(const char *);

/* music_midi.c */
extern int music_midi_process(void);

/* music_mp3.c */
extern int music_mp3_process(void);

/* menu.c */
extern void MenuPopup(Widget, XEvent *, String *, Cardinal *);
extern void CreateMenu(XtAppContext, Widget);
extern void load_dialog(void);

/* png.c */
extern void output_omake(void);
extern void capture_image(char *);

/* store_game.c */
extern int save_game(const char *);
extern int load_game(const char *);
extern int put4b(FILE *, int);
extern int get4b(FILE *);

/* store_global.c */
int save_global(const char *);
int load_global(const char *);

#undef EXTERN
#ifdef MAIN
#define EXTERN

int res2track[] = {
   0, -1,  2,  3,  4,  5,  6,  7,  8, 29, 
  -1, 27, -1, 34,  9, 10, 11, 26, 12, 13,
  28, 14, 30, 32, 15, 33, 16, 17, 30, 18,
  19, 31, -1, 25, -1, -1, 20, 21, 22, 23,
  24, 				1
};

#else
#define EXTERN extern

extern int res2track[];
#endif

EXTERN ARCHIVE cg, msg;
EXTERN PATH path;
EXTERN SCN_STAT scn_stat;
EXTERN SYS_STAT sys_stat;
EXTERN Widget top, core;

#endif
