/*
 * Hanashizume no Matsuri for X
 * Copyright (c) 2000 TF <tf@denpa.org>
 * All rights reserverd.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgment:
 *      This product includes software developed by TF.
 * 4. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: music_cd.c,v 1.15 2000/08/12 10:32:56 tf Exp $
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/wait.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/time.h>

#include "hana.h"

#ifndef True
#include <X11/Xlib.h>
#endif

typedef struct {
  int minute;
  int second;
  int frame;
} MSB;

typedef struct {
  MSB head;
  MSB end;
} CDTimeInfo;

typedef struct {
  int fd;
  int track_num;
  CDTimeInfo *tinfo;
} CDInfo;

typedef struct {
  MSB current;
  int status;
} CDStatus;

int music_cd_process(const char *);
CDInfo *music_cd_open(const char *);
int music_cd_play(CDInfo *, int);
int music_cd_start(CDInfo *);
int music_cd_stop(CDInfo *);
void music_cd_stop_fadeout(CDInfo *, int);
int music_cd_pause(CDInfo *);
int music_cd_resume(CDInfo *);
int music_cd_get_current_status(CDInfo *, CDStatus *);
void music_cd_get_volume(CDInfo *, int *);
int music_cd_set_volume(CDInfo *, int);
CDInfo *music_cd_open(const char *);
void music_cd_close(CDInfo *);
void timer_func(int);
static void cd_wait_child(void);

Bool loop;
CDInfo *c;
int track;

#define INTERVAL 100000

#if 0
int main(int argc, char *argv[])
{
  int fd;
  char buf[BUFSIZ];
  int i;

  fd =  music_cd_process(path.cddevice);
  while(1) {
    i = read(0, buf, BUFSIZ - 1);
    write(fd, buf, i);
  }
}
#endif

int music_cd_process(const char *cddevice)
{
  FILE *fp;
  char buf[BUFSIZ];
  int arg;
  static Bool pause_loop;
  struct itimerval tval;
  int resid;

  int fd[2];
  pid_t pid;

#ifndef AF_LOCAL
#define AF_LOCAL AF_UNIX
#define PF_LOCAL PF_UNIX
#endif
  if (socketpair(AF_LOCAL, SOCK_STREAM, PF_LOCAL, fd) < 0) {
    perror("socketpair AF_LOCAL/PF_LOCAL");
    if (socketpair(AF_UNIX, SOCK_STREAM, AF_UNSPEC, fd) < 0) {
      perror("socketpair AF_LOCAL/AF_UNSPEC");
      return -1;
    }
  } 

  if ((pid = fork()) < 0) {
    perror("fork");
    close(fd[0]);
    close(fd[1]);
    return -1;
  } else if (pid) {
    signal(SIGCHLD, (void *)cd_wait_child);
    close(fd[1]);
    return  fd[0];
  }
  
  fprintf(stderr, "started CD-DA process.\n");
  close(fd[0]);

  /* set Alarm procedure to check the CD status */
  loop = False;
  signal(SIGALRM, timer_func);
  
  tval.it_interval.tv_sec = 0;
  tval.it_interval.tv_usec = INTERVAL;
  tval.it_value.tv_sec = 0;
  tval.it_value.tv_usec = INTERVAL;
  
  setitimer(ITIMER_REAL, &tval, NULL);
  
  c = music_cd_open(cddevice);
  
  if (!c) {
    fprintf(stderr, "can't open %s.\n", cddevice);
  }
  
  if ((fp = fdopen(fd[1], "rw")) == NULL) {
    perror("fdopen");
    exit(1);
  }

  while (fgets(buf, sizeof(buf), fp)) {
    char *p;
    if ((p = strchr(buf, '\n'))) {
      *p = '\0';
    }
    switch(buf[0]) {
    case 't': /* play track */
      sscanf(buf + 1, "%d", &resid);
      track = res2track[resid];
      if (track < 0) {
	fprintf(stderr, "Invalid music id: %d.\n", resid);
	break;
      }
      if (music_cd_play(c, track) < 0) {
	fprintf(stderr, "can't play track %d.\n", track);
      }
      loop = False;
      break;
    case 'l': /* play track with loop */
      sscanf(buf + 1, "%d", &resid);
      track = res2track[resid];
      if (track < 0) {
	fprintf(stderr, "Invalid music id: %d.\n", resid);
	loop = False;
	break;
      }
      /* fprintf(stderr, "play track %d[id: %d] with loop\n", track, resid); */
      if (music_cd_play(c, track) < 0) {
	fprintf(stderr, "can't play track %d.\n", track);
	loop = False;
	break;
      }
      loop = True;
      break;
    case 'p': /* pause */
      music_cd_pause(c);
      pause_loop = loop;
      loop = False;
      break;
    case 'r': /* resume*/
      music_cd_resume(c);
      loop = pause_loop;
      break;
    case 's': /* stop */
      music_cd_stop(c);
      loop = False;
      break;
    case 'f': /* fadeout */
      sscanf(buf + 1, "%d", &arg);
      printf("fadeout in %d msec.\n", arg);
      music_cd_stop_fadeout(c, arg);
      break;
    case 'v': /* set volume */
      sscanf(buf + 1, "%d", &arg);
      printf("set volume to %d.\n", arg);
      music_cd_set_volume(c, arg);
      break;
    case 'x': /* exit */
      goto exit;
    }
  }
exit:
  music_cd_close(c);
  exit(0);
}

void timer_func(int sig)
{
  CDStatus s;

  if (!loop) {
    return;
  }

  if (music_cd_get_current_status(c, &s) < 0) {
    // fprintf(stderr, "loop.\n");
    if (music_cd_play(c, track) < 0) {
      fprintf(stderr, "can't play track %d.\n", track);
    }
  }
}

#ifdef __linux__
#include <linux/cdrom.h>

static int music_cd_get_entry(CDInfo *info)
{
  struct cdrom_tochdr toc_header;
  struct cdrom_tocentry toc;
  int i;

  /* clear the old info */
  if (info->tinfo) {
    free(info->tinfo);
  }
  if (ioctl(info->fd, CDROMREADTOCHDR, &toc_header) < 0) {
    return -1;
  }
  
  /* get header info */
  info->track_num = toc_header.cdth_trk1;
  
  info->tinfo = (CDTimeInfo *)calloc(toc_header.cdth_trk1, sizeof(CDTimeInfo));
  /* get TOC */
  toc.cdte_format = CDROM_MSF;
  for (i = 1; i <= toc_header.cdth_trk1; i++) {
    toc.cdte_track = i;
    if (ioctl(info->fd, CDROMREADTOCENTRY, &toc) < 0) {
      return -1;
    }
    info->tinfo[i - 1].head.minute = toc.cdte_addr.msf.minute;
    info->tinfo[i - 1].head.second = toc.cdte_addr.msf.second;
    info->tinfo[i - 1].head.frame = toc.cdte_addr.msf.frame;
    if (i >= 2) {
      info->tinfo[i - 2].end.minute = toc.cdte_addr.msf.minute;
      info->tinfo[i - 2].end.second = toc.cdte_addr.msf.second;
      info->tinfo[i - 2].end.frame = toc.cdte_addr.msf.frame;
    }
  }
  toc.cdte_track = CDROM_LEADOUT;
  if (ioctl(info->fd, CDROMREADTOCENTRY, &toc) < 0) {
    return -1;
  }
  info->tinfo[i - 2].end.minute = toc.cdte_addr.msf.minute;
  info->tinfo[i - 2].end.second = toc.cdte_addr.msf.second;
  info->tinfo[i - 2].end.frame = toc.cdte_addr.msf.frame;
  
  for (i = 0; i < toc_header.cdth_trk1; i++) {
    printf("%5d %02d:%02d.%03d - %02d:%02d.%03d [%c]\n",
	   i + 1,
	   info->tinfo[i].head.minute,
	   info->tinfo[i].head.second,
	   info->tinfo[i].head.frame,
	   info->tinfo[i].end.minute,
	   info->tinfo[i].end.second,
	   info->tinfo[i].end.frame,
	   (toc.cdte_ctrl & CDROM_DATA_TRACK ? 'D' : 'A'));
  }
  return 0;
}

int music_cd_play(CDInfo *info, int track)
{
  struct cdrom_msf msf;

  if (!info) {
    return 0;
  }

  if (track < 1 || track > info->track_num) {
    fprintf(stderr, "Invalid track %d is specified.\n", track);
    return -1;
  }
  track--;

  msf.cdmsf_min0 = info->tinfo[track].head.minute;
  msf.cdmsf_sec0 = info->tinfo[track].head.second;
  msf.cdmsf_frame0 = info->tinfo[track].head.frame;
  
  msf.cdmsf_min1 = info->tinfo[track].end.minute;
  msf.cdmsf_sec1 = info->tinfo[track].end.second;
  msf.cdmsf_frame1 = info->tinfo[track].end.frame;

#if 0
  printf("play track %d -> %02d:%02d.%03d - %02d:%02d.%03d\n",
	 track + 1,
	 msf.cdmsf_min0,
	 msf.cdmsf_sec0,
	 msf.cdmsf_frame0,
	 msf.cdmsf_min1,
	 msf.cdmsf_sec1,
	 msf.cdmsf_frame1);
#endif
  
  return ioctl(info->fd, CDROMPLAYMSF, &msf);
}

int music_cd_start(CDInfo *info)
{
  if (info) {
    return ioctl(info->fd, CDROMSTART);
  } else {
    return 0;
  }
}

int music_cd_stop(CDInfo *info)
{
  if (info) {
    return ioctl(info->fd, CDROMSTOP);
  } else {
    return 0;
  }
}

void music_cd_stop_fadeout(CDInfo *info, int msec)
{
  int vol, orig_vol, step, vol_step, i;

#define FADE_STEP 100

  if (!info) {
    return;
  }

  music_cd_get_volume(info, &orig_vol);
  fprintf(stderr, "vol: %d\n", orig_vol);
  
  step = msec / FADE_STEP;
  vol = orig_vol;
  vol_step = vol / step;
  
  for (i = 0; i < step; i++) {
    vol -= vol_step;
    music_cd_set_volume(info, vol);
    usleep(FADE_STEP * 1000);
  }
  
  music_cd_set_volume(info, orig_vol);
  ioctl(info->fd, CDROMSTOP);
}

int music_cd_pause(CDInfo *info)
{
  if (info) {
    return ioctl(info->fd, CDROMPAUSE);
  } else {
    return 0;
  }
}

int music_cd_resume(CDInfo *info)
{
  if (info) {
    return ioctl(info->fd, CDROMRESUME);
  } else {
    return 0;
  }
}

int music_cd_get_current_status(CDInfo *info, CDStatus *status)
{
  struct cdrom_subchnl s;
  
  if (!info) {
    return -1;
  }
  
  s.cdsc_format = CDROM_MSF;
  if (ioctl(info->fd, CDROMSUBCHNL, &s) < 0) {
    return -1;
  }
  if (s.cdsc_audiostatus != CDROM_AUDIO_PLAY) {
    return -1;
  }
  
  if (status) {
    status->current.minute = s.cdsc_reladdr.msf.minute;
    status->current.second = s.cdsc_reladdr.msf.second;
    status->current.frame = s.cdsc_reladdr.msf.frame;
  }
  
  return 0;
}

void music_cd_get_volume(CDInfo *info, int *vol)
{
  struct cdrom_volctrl v;

  if (info) {
    ioctl(info->fd, CDROMVOLCTRL, &v);
    *vol = v.channel0;
  }
}

int music_cd_set_volume(CDInfo *info, int vol)
{
  struct cdrom_volctrl v;

  if (info) {
    v.channel0 = vol;
    v.channel1 = vol;
    v.channel2 = vol;
    v.channel3 = vol;
    return ioctl(info->fd, CDROMVOLCTRL, &v);
  } else {
    return 0;
  }
}

CDInfo *music_cd_open(const char *devicename)
{
  CDInfo *c;
  int fd;

  if ((fd = open(devicename, O_RDONLY, 0)) < 0) {
    return NULL;
  }

  if ((c = malloc(sizeof *c)) == NULL) {
    perror("music_cd_open");
    exit(1);
  }
  
  c->fd = fd;
  c->track_num = 0;
  c->tinfo = NULL;
  
  if (music_cd_get_entry(c)) {
    music_cd_close(c);
    return NULL;
  }
  music_cd_start(c);
  
  return c;
}

void music_cd_close(CDInfo *info)
{
  if (info) {
    music_cd_stop(info);
    close(info->fd);
    if (info->tinfo) {
      free(info->tinfo);
    }
    free(info);
  }
}
#elif BSD
#include <sys/cdio.h>

static u_int msf2lba(u_char, u_char, u_char);

static u_int msf2lba(u_char m, u_char s, u_char f)
{
        return (((m * 60) + s) * 75 + f) - 150;
}

static int music_cd_get_entry(CDInfo *info)
{
  struct ioc_toc_header toc_header;
  struct ioc_read_toc_entry toc;
  int i;

  /* clear the old info */
  if (info->tinfo != NULL) {
    free(info->tinfo);
  }
  if (ioctl(info->fd, CDIOREADTOCHEADER, &toc_header) < 0) {
    return -1;
  }
  
  /* get header info */
  info->track_num = toc_header.ending_track - toc_header.starting_track + 1;
  
  /* get TOC */
  toc.address_format = CD_MSF_FORMAT;
  toc.starting_track = 0;

  toc.data_len = sizeof(*toc.data) * (info->track_num + 1);
  toc.data = malloc(toc.data_len);
  if (toc.data == NULL) {
    perror("music_cd_get_entry");
    exit(1);
  }

  if (ioctl(info->fd, CDIOREADTOCENTRYS, (char *)&toc) < 0) {
    free(toc.data);
    return -1;
  }

  info->tinfo = malloc(sizeof(info->tinfo[0]) * info->track_num);
  if (info->tinfo == NULL) {
    perror("cdinfo_get_entry");
    exit(1);
  }

  for (i = 0; i < info->track_num; i++) {
    info->tinfo[i].head.minute = toc.data[i].addr.msf.minute;
    info->tinfo[i].head.second = toc.data[i].addr.msf.second;
    info->tinfo[i].head.frame = toc.data[i].addr.msf.frame;
    if (i >= 1) {
      info->tinfo[i-1].end.minute = toc.data[i].addr.msf.minute;
      info->tinfo[i-1].end.second = toc.data[i].addr.msf.second;
      info->tinfo[i-1].end.frame = toc.data[i].addr.msf.frame;
    }
  }
  info->tinfo[i-1].end.minute = toc.data[i].addr.msf.minute;
  info->tinfo[i-1].end.second = toc.data[i].addr.msf.second;
  info->tinfo[i-1].end.frame = toc.data[i].addr.msf.frame;

#ifdef DEBUG
  for (i = 0; i < info->track_num; i++) {
    printf("%5d %02d:%02d.%03d - %02d:%02d.%03d [%s]\n",
	   i + 1,
	   info->tinfo[i].head.minute,
	   info->tinfo[i].head.second,
	   info->tinfo[i].head.frame,
	   info->tinfo[i].end.minute,
	   info->tinfo[i].end.second,
	   info->tinfo[i].end.frame,
#define	CDROM_DATA_TRACK	0x04
	   (toc.data[i].control & CDROM_DATA_TRACK ? "DATA" : "AUDIO"));
  }
#endif

  free(toc.data);
  return 0;
}

int music_cd_play(CDInfo *info, int track)
{
  struct ioc_play_blocks t;

  if (info == NULL) {
    return 0;
  }

  if (track < 1 || track > info->track_num) {
    fprintf(stderr, "Invalid track %d is specified.\n", track);
    return -1;
  }

  t.blk = msf2lba(info->tinfo[track-1].head.minute,
                  info->tinfo[track-1].head.second,
		  info->tinfo[track-1].head.frame);
  t.len = msf2lba(info->tinfo[track].head.minute,
                  info->tinfo[track].head.second,
		  info->tinfo[track].head.frame)
		      - t.blk;

#ifdef DEBUG
  printf("play track %d -> %02d:%02d.%03d - %02d:%02d.%03d\n",
	 track,
	 info->tinfo[track-1].head.minute,
	 info->tinfo[track-1].head.second,
	 info->tinfo[track-1].head.frame,
	 info->tinfo[track].head.minute,
	 info->tinfo[track].head.second,
	 info->tinfo[track].head.frame);
#endif
  
  return ioctl(info->fd, CDIOCPLAYBLOCKS, &t);
}

int music_cd_start(CDInfo *info)
{
  if (info == NULL) {
    return 0;
  }
  return ioctl(info->fd, CDIOCSTART);
}

int music_cd_stop(CDInfo *info)
{
  if (info == NULL) {
    return 0;
  }
  return ioctl(info->fd, CDIOCSTOP);
}

int music_cd_pause(CDInfo *info)
{
  if (info == NULL) {
    return 0;
  }
  return ioctl(info->fd, CDIOCPAUSE);
}

int music_cd_resume(CDInfo *info)
{
  if (info == NULL) {
    return 0;
  }
  return ioctl(info->fd, CDIOCRESUME);
}

void music_cd_stop_fadeout(CDInfo *info, int msec)
{
  int vol, orig_vol, step, vol_step, i;

  if (info == NULL) {
    return;
  }

  music_cd_get_volume(info, &orig_vol);
#ifdef DEBUG
  fprintf(stderr, "vol: %d\n", orig_vol);
#endif
  
#define FADE_STEP 100
  step = msec / FADE_STEP;
  vol = orig_vol;
  vol_step = vol / step;
  
  for (i = 0; i < step; i++) {
    vol -= vol_step;
    music_cd_set_volume(info, vol);
    usleep(FADE_STEP * 1000);
  }
  
  music_cd_set_volume(info, orig_vol);
  music_cd_stop(info);
}

int music_cd_get_current_status(CDInfo *info, CDStatus *status)
{
  struct ioc_read_subchannel s;
  struct cd_sub_channel_info data;
  
  if (info == NULL) {
    return -1;
  }

  memset(&s, 0, sizeof(s));
  s.data = &data;
  s.data_len = sizeof(data);
  s.data_format = CD_CURRENT_POSITION;
  s.address_format = CD_MSF_FORMAT;

  if (ioctl(info->fd, CDIOCREADSUBCHANNEL, (char *)&s) < 0) {
    return -1;
  }
  if (data.header.audio_status != CD_AS_PLAY_IN_PROGRESS) {
    return -1;
  }

  if (status) {
    status->current.minute = s.data->what.position.reladdr.msf.minute;
    status->current.second = s.data->what.position.reladdr.msf.second;
    status->current.frame = s.data->what.position.reladdr.msf.frame;
  }
  
  return 0;
}

void music_cd_get_volume(CDInfo *info, int *vol)
{
  struct ioc_vol v;

  if (info != NULL) {
    ioctl(info->fd, CDIOCGETVOL, &v);
    *vol = v.vol[0];
  }
}

int music_cd_set_volume(CDInfo *info, int vol)
{
  struct ioc_vol v;

  if (info == NULL) {
    return 0;
  }
 
  v.vol[0] = vol;
  v.vol[1] = vol;
  v.vol[2] = vol;
  v.vol[3] = vol;
  return ioctl(info->fd, CDIOCSETVOL, &v);
}

CDInfo *music_cd_open(const char *devicename)
{
  CDInfo *c;
  int fd;

  if ((fd = open(devicename, O_RDONLY, 0)) < 0) {
    return NULL;
  }

  if ((c = malloc(sizeof *c)) == NULL) {
    perror("music_cd_open");
    exit(1);
  }
  
  c->fd = fd;
  c->track_num = 0;
  c->tinfo = NULL;
  
  if (music_cd_get_entry(c)) {
    music_cd_close(c);
    return NULL;
  }
  music_cd_start(c);
  
  return c;
}

void music_cd_close(CDInfo *info)
{
  if (info != NULL) {
    music_cd_stop(info);
    close(info->fd);
    if (info->tinfo != NULL) {
      free(info->tinfo);
    }
    free(info);
  }
}
#else
int music_cd_play(CDInfo *info, int track)
{
  return -1;
}

int music_cd_start(CDInfo *info)
{
  return -1;
}

int music_cd_stop(CDInfo *info)
{
  return -1;
}

void music_cd_stop_fadeout(CDInfo *info, int msec)
{
}

int music_cd_pause(CDInfo *info)
{
  return -1;
}

int music_cd_resume(CDInfo *info)
{
  return -1;
}

int music_cd_get_current_status(CDInfo *info, CDStatus *status)
{
  return -1;
}

void music_cd_get_volume(CDInfo *info, int *vol)
{
}

int music_cd_set_volume(CDInfo *info, int vol)
{
  return -1;
}

CDInfo *music_cd_open(const char *devicename)
{
  return NULL;
}

void music_cd_close(CDInfo *info)
{
}
#endif

static void cd_wait_child()
{
  int pid, status;

  pid = wait(&status);
}
